from scipy import stats
import copy

from ml_processor.configuration import config 

class remove_outliers:

    """

    Removing outliers from data.
    
    Parameters
    ----------
        
    data : pandas.DataFrame
        Data set from which to remove outliers

    columns : list or arraay-like 
        Names of columns from which to remove outliers
        
    split :boolean (default=True) 
        Whether to remove outliers according to different classes of a target column
        
    target : string (default='target') 
        Column name with classes for removing outliers separately if split=True

    log : boolean (default=False)
        Whether to print out information on the results

    """
    def __init__(self, data, columns, split=True, target='target', log=False):
        
        self.data = data
        
        self.columns = columns
        
        self.split = split
        
        self.target = target
        
        self.log=log
        
        self.logger = config.get_logger()
    
    def std_method(self, threshold=3):

        """

        Remove outliers from data set using the z-score method.
        
        Parameters
        ----------
        
        threshold : int (default=3) 
            z-score from which the value is an outlier
            
        Returns
        -------
        
        pandas.DataFrame: 
            Dataset of original dataset without outliers

        """
        
        data = self.data.copy()
        
        if self.split:
            
            for col in self.columns:
                
                data = data.groupby(self.target).apply(lambda x: x[(abs(stats.zscore(x[col])) < threshold)])
                
                data.index = data.index.droplevel(0)
            
            if self.log:
                
                labels = self.data[self.target].unique()
                
                for col in self.columns:
                    
                    mask = []
                    
                    for label in labels:
                        
                        check_data = self.data[self.data[self.target]==label]
                        
                        rm_index = list(check_data[(np.abs(stats.zscore(check_data[col])) >= threshold)].index)
                        
                        mask.extend(rm_index)
                        
                        rows_dropped = len(rm_index)
                        
                        self.logger.info(f'Number of rows dropped for column={col} and target_label={label}: {rows_dropped}')
                        
                        floor = self.data.loc[~self.data.index.isin(rm_index)][col].min()
                        
                        ceiling = self.data.loc[~self.data.index.isin(rm_index)][col].max()
                        
                        self.logger.info(f'Data range for column={col} and target_label={label}: {[floor, ceiling]}')
                    
                    total_rows_dropped = len(set(mask))
                    
                    pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))
                    
                    self.logger.info(f'Number of rows dropped for column={col}: {total_rows_dropped} ({pct_dropped})')
        else:
            
            mask = []
            
            for col in self.columns:
                
                data = data[(abs(stats.zscore(data[col])) < threshold)]
            
            if self.log:
                
                for col in self.columns:
                    
                    rm_index = list(self.data[(abs(stats.zscore(self.data[col])) < threshold)].index)
                    
                    total_rows_dropped = len(self.data) - len(rm_index)
                    
                    pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))
                    
                    self.logger.info(f'Number of rows dropped for column={col}: {total_rows_dropped} ({pct_dropped})')
    
        if self.log:
            
            total_rows_dropped = len(self.data) - len(data)
            
            pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))
            
            self.logger.info(f'Total number of rows dropped: {total_rows_dropped} ({pct_dropped})')

        return data
    
    def percentile_method(self, threshold=0.95):

        """

        Remove outliers from data set using the percentile method.
        
        Parameters
        ----------
        
        threshold : int (default=0.95) 
            Percentile from which the value is an outlier
            
        Returns
        -------
        
        pandas.DataFrame: 
            Dataset of original dataset without outliers

        """
        
        data = self.data.copy()
        
        if self.split:
            
            columns = self.columns + [self.target]
            
            quantiles = self.data[columns].groupby(self.target).quantile(threshold)
            
            mask_all = []
            
            for col in self.columns:
                
                label_qtiles = quantiles[col].to_dict()
                
                mask = []
                
                for label in label_qtiles:
                    
                    check_data = self.data[(self.data[self.target]==label)]
                    
                    rm_index = list(check_data[check_data[col] >= label_qtiles[label]].index)
                    
                    mask.extend(rm_index)
                    
                    rows_dropped = len(rm_index)
                    
                    if self.log:
                        
                        self.logger.info(f'Number of rows dropped for column={col} and target_label={label}: {rows_dropped}')
                        
                        self.logger.info(f'Ceiling for column={col} and target_label={label}: {label_qtiles[label]}')
                
                if self.log:
                    
                    total_rows_dropped = len(set(mask))
                    
                    pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))
                    
                    self.logger.info(f'Number of rows dropped for column={col}: {total_rows_dropped} ({pct_dropped})')             
                
                mask_all.extend(mask) 
            
            data = data.loc[~data.index.isin(mask_all)]
            
            if self.log:

                total_rows_dropped = len(set(mask_all))

                pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))

                self.logger.info(f'Total number of rows dropped: {total_rows_dropped} ({pct_dropped})')

        else:
            
            quantiles = self.data[self.columns].quantile(threshold)
            
            cut_offs = zip(quantiles.index, quantiles.values)
            
            mask_all = []
            
            for (col, cut_off) in cut_offs:
                
                rm_index = list(self.data[self.data[col] >= cut_off].index)
                
                mask_all.extend(rm_index)
                
                if self.log:
                    
                    rows_dropped = len(rm_index)
                    
                    pct_dropped = '{:.1%}'.format(rows_dropped/len(self.data))
                    
                    self.logger.info(f'Number of rows dropped for column={col}: {rows_dropped} ({pct_dropped})')
                    
            data = self.data.loc[~self.data.index.isin(mask_all)]
        
            if self.log:

                total_rows_dropped = len(set(mask_all))

                pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))

                self.logger.info(f'Total number of rows dropped: {total_rows_dropped} ({pct_dropped})')
            
        return data
    
    
    def iqr_method(self):

        """

        Remove outliers from data set using the inter-quantile method.
        
        Parameters
        ----------
        
        None
            
        Returns
        -------
        
        pandas.DataFrame: 
            Dataset of original dataset without outliers

        """

        data = self.data.copy()
        
        if self.split:
            
            mask_all = []
            
            for col in self.columns:
                
                mask = []
                
                labels = self.data[self.target].unique()
                
                for label in labels:
                    
                    check_data = self.data[(self.data[self.target]==label)]

                    Q1, Q3 = check_data[col].quantile([0.25, 0.75])
                    
                    IQR = Q3 - Q1
                    
                    ll = Q1 - (1.5 * IQR)
                    
                    ul = Q1 + (1.5 * IQR)

                    rm_index = list(check_data[(check_data[col] >= ul) | (check_data[col] <= ll)].index)
                    
                    mask.extend(rm_index)
                    
                    if self.log:

                        rows_dropped = len(rm_index)

                        pct_dropped = '{:.1%}'.format(rows_dropped/len(self.data))

                        self.logger.info(f'Number of rows dropped for column={col}: {rows_dropped} ({pct_dropped})')
                        
                        self.logger.info(f'Data range for column={col} and target_label={label}: {[ll, ul]}')
                
                mask_all.extend(mask)
                
                if self.log:
                    
                    total_rows_dropped = len(set(mask))
                    
                    pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))
                    
                    self.logger.info(f'Number of rows dropped for column={col}: {total_rows_dropped} ({pct_dropped})')  
            
            data = self.data.loc[~self.data.index.isin(mask_all)]
            
            if self.log:

                total_rows_dropped = len(set(mask_all))

                pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))

                self.logger.info(f'Total number of rows dropped: {total_rows_dropped} ({pct_dropped})')
        
        else:
            
            mask_all = []
            
            for col in self.columns:
                
                Q1, Q3 = self.data[col].quantile([0.25, 0.75])
                
                IQR = Q3 - Q1
                
                ll = Q1 - (1.5 * IQR)
                
                ul = Q1 + (1.5 * IQR)
                
                rm_index = list(self.data[(self.data[col] >= ul) | (self.data[col] <= ll)].index)
                
                mask_all.extend(rm_index)
                
                if self.log:
                    
                    rows_dropped = len(rm_index)

                    pct_dropped = '{:.1%}'.format(rows_dropped/len(self.data))

                    self.logger.info(f'Number of rows dropped for column={col}: {rows_dropped} ({pct_dropped})')

                    self.logger.info(f'Data range for column={col}: {[ll, ul]}')
            
            data = self.data.loc[~self.data.index.isin(mask_all)]
            
            if self.log:

                total_rows_dropped = len(set(mask_all))

                pct_dropped = '{:.1%}'.format(total_rows_dropped/len(self.data))

                self.logger.info(f'Total number of rows dropped: {total_rows_dropped} ({pct_dropped})')

        return data




import logging
import sys
# from dotenv import dotenv_values 
import json
import numpy as np
import os



class config: 
        
    # def add_path(lib_path=None):

    #     """

    #     Add path to working path

    #     Parameters
    #     ----------

    #     lib_path : string (default=None)
    #         Path to add to working path

    #     Returns
    #     -------

    #     None

    #     """
    #     if lib_path:
    #         sys.path.insert(1, lib_path)
    #     else:
    #         lib_path = os.path.join(os.environ.get("HOME"), "Library/CloudStorage/OneDrive-MoneseLtd/GW/libraries/snowflake")
    #         sys.path.insert(1, lib_path)  
    
    # def get_credentials(env_path=None):

    #     """
        
    #     Get credentials stored in dot file

    #     Parameters
    #     ----------
        
    #     env_path : string (deault=None)
    #         Path to dot file with credentials

    #     Returns
    #     -------

    #     Dic:
    #         Stored credentials

    #     """
        
    #     if env_path:
    #         env_path = env_path
    #     else:
    #         env_path = os.path.join(os.environ.get("HOME"), "Library/CloudStorage/OneDrive-MoneseLtd/GW/libraries/environ_vars/.env")
            
    #     return dotenv_values(env_path)
    
    def get_logger(file=None, folder="model_artefacts"):

        """

        Create logger

        Parameters
        ----------

        file : string (default=None)

        Returns
        -------

        object:
            Logger

        """

        logger = logging.getLogger('logger_name')

        logger.setLevel(logging.INFO)

        logger.handlers = []

        formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(message)s")

        streamhandler = logging.StreamHandler(stream=sys.stdout)

        streamhandler.setLevel(logging.DEBUG)

        streamhandler.setFormatter(formatter)

        logger.addHandler(streamhandler)

        if file:

            if not os.path.isdir(f'{ folder}'):
                
                os.mkdir(f'{folder}')

            path = os.path.join(f'{folder}', file)

            filehandler = logging.FileHandler(path)

            filehandler.setLevel(logging.DEBUG)

            filehandler.setFormatter(formatter)

            logger.addHandler(filehandler)

        logger.propagate=False

        return logger


        



import json 

import numpy as np

class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



import pandas as pd
import numpy as np
from sklearn.preprocessing import OneHotEncoder
from optbinning import BinningProcess
import os
import pickle
import warnings
import time

import seaborn as sns
import matplotlib.pyplot as plt

warnings.filterwarnings('ignore')
sns.set_style('whitegrid')

from ml_processor.configuration import config

# <<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
class data_prep:

    """
    Data preparation to transform data using one hot encoding or woe transformation.

    Parameters
    ----------
    data : pandas.DataFrame 
        Data set to transform

    features : list or array-like
        Names of columns (variables) to transform using either one hot encoding or woe transformation

    target : string
        Name of the column with binary labels

    categories : list or array-like
        Names of categorical variables in the dataset. 


    """
    
    def __init__(self, data, features, target='target', categories=None):
        
        self.data = data
        
        self.features = features
        
        self.categories = categories
        
        self.target = target

        self.log = config.get_logger()

        if not os.path.isdir('data_prep'):
            os.mkdir('data_prep')
        
        self.binning_process = None
        
        self.encoder = None

        self.optb = None
        
    def create_encoder(self, drop='if_binary', verbose=False):

        """

        Create encode for one hot encoding

        Parameters
        ----------

        drop : string (default='if_binary', possible values {'first', 'if_binary'} or an array-like of shape (n_features,))  
            Specifies a methodology to use to drop one of the categories per feature.

        verbose : Boolean (default=False)
            Log to console or not.

        Returns
        -------
        
        None

        """

        start = int(time.time())
        
        X = self.data[self.categories].values
        
        self.encoder = OneHotEncoder(drop=drop, sparse=False)
        
        self.encoder = self.encoder.fit(X)

        end = int(time.time())

        if verbose:
            self.log.info('OneHotEncoder created')

        if verbose:
            self.log.info(f'Runtime for creating encoder: {int(end-start)} seconds')

        path = os.path.join(os.getcwd(), 'data_prep', 'encoder')

        with open(path, "wb") as file: 
            pickle.dump(self.encoder, file)

        if verbose:
            self.log.info(f'OneHot encoder saved: {path}')
        
    def oneHot_transform(self, data=pd.DataFrame(), verbose=False):

        """

        Transform data using one hot encoding

        Parameters
        ----------    
        data : pandas.DataFrame (default=empty dataframe)    
            Data to transfrom using one hot encoding

        Returns
        -------
        df_transformed : pandas.DataFrame
            Transformed dataset
            
        """
        
        if len(data) == 0:
            data = self.data

        X = data[self.categories].values

        if not self.encoder:
            self.create_encoder()

        start = int(time.time())

        X_encoded = self.encoder.transform(X)

        end = int(time.time())

        df_encoded = pd.DataFrame(X_encoded, columns=self.encoder.get_feature_names(self.categories))

        if verbose:
            self.log.info('Categorical variables succesful encoded')

        if verbose:
            self.log.info(f'Runtime for encoding categorical variables: {int(end-start)} seconds')

        df_transformed = pd.concat([data.reset_index(drop=True), df_encoded], axis=1)

        features = [col for col in self.features if col not in self.categories]

        self.features = list(set(features + list(df_encoded.columns)))

        if verbose:
            self.log.info(f'Total number of features after encoding: {len(self.features)}')
        
        self.df_transformed =  df_transformed

        
        return df_transformed
        
    def woe_bins(self, min_prebin_size = 0.1, 
                 selection_criteria = {"iv": {"min": 0.01, "max": 0.7, "strategy": "highest", "top": 50},
                                       "quality_score": {"min": 0.01}},
                 binning_fit_params = None,
                 verbose=False,
                 ):

        """

        Generate binning process for woe transformation
        
        Parameters
        ----------    
        
        min_prebin_size : float (default=0.1)
            The fraction of minimum number of records for each bin
            
        selection_criteria : dict or None 
            default
            -------
            {"iv": {"min": 0.001, "max": 0.7, "strategy": "highest", "top": 50}, 
            "quality_score": {"min": 0.001}}
            
            Variable selection criteria

        binning_fit_params : dict or None (default=None)
            Dictionary with optimal binning transform options for specific
            variables. Example ``{"variable_1": {"metric": "event_rate"}}``.
            
        Returns
        -------

        Self - Fitted binning process.
        
        """ 

        start = int(time.time())

        X = self.data[self.features]
        
        y = self.data[self.target]
        
        binning_process = BinningProcess(
            self.features,
            categorical_variables = self.categories,
            min_prebin_size = min_prebin_size,
            selection_criteria=selection_criteria,
            binning_fit_params = binning_fit_params
        )
        
        binning_process.fit(X, y)

        end = int(time.time())

        if verbose:
            self.log.info(f'Binning processor created')

        if verbose:
            self.log.info(f'Runtime for creating binning process : {int(end-start)} seconds')
        
        self.binning_process = binning_process

        path = os.path.join(os.getcwd(), 'data_prep', 'binningprocess.pkl')

        binning_process.save(path)

        if verbose:
            self.log.info(f'Binning processor saved : {path}')
    
    def woe_bin_table(self):

        """

        Generate summary results for the binning process
        
        Parameters
        ---------- 

        None
        
        Returns
        -------
        
        woe_table : pandas.DataFrame
            Summary results on the binning process
        
        """
        
        if not self.binning_process:
            self.woe_bins()
        
        woe_table = self.binning_process.summary()
        
        woe_table = woe_table.sort_values('iv', ascending=False).reset_index(drop=True)
        
        return woe_table
    
    def get_var_bins(self, var, ax, plot_type="WoE"):

        """

        Generate binning details for the variable
        
        Parameters
        ---------- 
        
        var : string 
            Name of variable for which to show binning tables

        ax : matplotlib.axes
            Axes to plot on.

        plot_type : string (default="WoE", options=[""WoE", "event_rate"])
            Whether to plot WoE or event rate per bin
            
        Returns
        -------   
        plot : matplotlib plot
        
        """

        if not self.binning_process:
            self.woe_bins()  
            
        self.optb = self.binning_process.get_binned_variable(var)

        table = self.optb.binning_table.build()
        
        table.drop(['Totals'], inplace=True)

        table['WoE'] = pd.to_numeric(table['WoE'])

        table = table.query("abs(WoE)>0")

        table['Bin'] = table['Bin'].astype('string')
        
        if plot_type == "WoE":

            sns.barplot(x='Bin', y='WoE', data=table, ax=ax, color='deepskyblue')

            ax.set_title('WoE per bins created \nfor {}'.format(var), fontsize = 12, fontweight='bold', fontname='Georgia')

            ax.tick_params(axis='both', which='major', labelsize=10)

            ax.tick_params(axis='x', which='major', rotation=90)

            ax.set_xlabel('Bin', fontsize=10)

            ax.set_ylabel('WoE', fontsize=10)

            ax.set_xlabel('Bin', fontsize=10)

            ax.spines["top"].set_visible(False)
            ax.spines["right"].set_visible(False)
            ax.spines["left"].set_visible(False)
        
        else:
            
            ax.bar(table['Bin'], table['Event'], color='#ff2e63', label='Event')
            
            ax.bar(table['Bin'], table['Non-event'], bottom=table['Event'], color='#48d1cc', label='Non-event')
            
            ax.tick_params(axis='both', which='both', labelsize=10)
            
            ax.tick_params(axis='x', which='major', rotation=90)
            
            ax.legend(frameon=False, fontsize=8)
            
            ax.set_ylabel('Count', fontsize=10)

            ax.set_title('Event rate per bin \nfor {}'.format(var), fontname='Georgia', fontsize=12, fontweight='bold')
            
            ax.set_xlabel('Bin', fontsize=10)

            ax.spines["top"].set_visible(False)
            ax.spines["right"].set_visible(False)
            ax.spines["left"].set_visible(False)

            
            table['Bin'] = table['Bin'].astype('category')
            
            ax2 = ax.twinx()
            
            ax2.plot(table['Bin'], table['Event rate'], color='k', marker='o', label='event rate')
            
            ax2.set_yticklabels(['{:.0%}'.format(x) for x in ax2.get_yticks()])
            
            ax2.set_ylabel('Event rate', fontsize=10)

            ax2.grid(False)

            ax2.tick_params(axis='y', which='both', right=False, left=False)

            ax2.spines["top"].set_visible(False)
            ax2.spines["right"].set_visible(False)
            ax2.spines["left"].set_visible(False)
        
    def woe_transform(self, data=pd.DataFrame(), verbose=False):

        """

        Transform data using Weight of Evidence (WOE) weights
        
        Parameters
        ----------
        
        data : pandas.DataFrame (default=empty dataframe)
            Dataset to tranform

        verbose : boolean (default=False)
            Log to console or not.
            
        Returns
        -------

        pandas.DataFrame
            Transformed data.
        
        """
        
        if len(data) == 0:
            data = self.data
        
        if not self.binning_process:
            self.woe_bins()

        start = int(time.time())
        
        data_transf = data[self.features]

        data_transf = self.binning_process.transform(data_transf, metric="woe")
        
        df_transformed = pd.DataFrame(data_transf)
        
        df_transformed['target'] = data[self.target].values  

        end = int(time.time())

        if verbose:
            self.log.info('Data transformation completed')

        if verbose:
            self.log.info(f'Runtime for woe encoding : {int(end-start)} seconds')
        
        self.df_transformed = df_transformed
        
        return df_transformed
    
    def woe_features(self, verbose=False):
        
        """

        Generate variables selected using the selection criteria
        
        Parameters
        ----------
        None
        
        Returns
        -------

        features : list
            variables selected
        
        """
        
        if not self.binning_process:
            self.woe_bins()
            
        features = self.binning_process.get_support(names=True)
        
        if verbose:
            self.log.info(f'Number of features selected using the selection criteria defined : {len(features)} out of {len(self.features)}')
        
        return features
    
    def balance_data(self, target=None, data=pd.DataFrame()):
        
        """

        Balance data basing on each label size of the label variable.
        
        Parameters
        ----------
        
        data : pandas.DataFrame (default=empty dataframe)
            Dataset to balance.
            
        Returns
        -------

        pandas.DataFrame
            Transformed data.
            
        """

        if len(data) == 0:
            data = self.data

        if not target:
            target = self.target
        
        minority_size = data[target].value_counts()[1]

        self.balanced_df = data.groupby(target).sample(minority_size)
        
        self.balanced_df = self.balanced_df.sample(frac=1)
        
        return self.balanced_df
        



import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt 
import seaborn as sns
import math
import re
import warnings

warnings.filterwarnings('ignore')
sns.set_style('whitegrid')

from ml_processor.configuration import config
from ml_processor.outliers import remove_outliers

def eda_data_quality(data, target=None):
    
    """

    Performing data quality checks on data set.
    
    Parameters
    ----------

    data : pandas.DataFrame 
        Data set to check.
        
    Returns
    -------

    df_results : pandas.DataFrame
        Results for quality checks

    """
    
    df_results = pd.DataFrame(data.dtypes)
    
    df_results.columns = ['type']
    
    df_results['type'] = df_results['type'].astype('string')
    
    df_uniq = pd.DataFrame(data.nunique())
    
    df_uniq.columns = ['unique']
    
    df_results = df_results.merge(df_uniq, how='left', left_index=True, right_index=True)
    
    missing = pd.DataFrame(data.isnull().sum())
    
    missing.columns = ['missing']
    
    df_results = df_results.merge(missing, how='left', left_index=True, right_index=True)
    
    df_results['pct.missing'] = (df_results['missing']/len(data))

    def get_mode(data):

        columns = data.columns.to_list()

        modes = {'column' : columns, 'mode' : [], 'pct.mode' : []}

        for col in columns:
            
            mode = data[col].value_counts(normalize=True).index[0]
            modes['mode'].append(mode)
            
            pct_mode = data[col].value_counts(normalize=True).values[0]
            modes['pct.mode'].append(pct_mode)
    
        mode_res = pd.DataFrame(modes)

        mode_res = mode_res.set_index('column')
        
        return mode_res
    
    df_modes = get_mode(data)

    df_results = df_results.merge(df_modes, how='left', left_index=True, right_index=True)
    
    df_summar = pd.DataFrame(data.describe()).T
    
    df_summar = df_summar[['mean', 'std', 'min', '25%', '50%', '75%', 'max']]
    
    df_results = df_results.merge(df_summar, how='left', left_index=True, right_index=True)
    
    df_results = df_results.drop('std', axis=1)
    
    def check_dtype(x, y):
        if (any(re.findall(r'float|int', x, re.IGNORECASE)) and y < 30) or (x=='object' and y>30):
            return 45
        else:
            return 0
        
    df_results = (
        df_results
        .assign(rule_1 = np.where(df_results['pct.missing'] > 0.5, 80, 0))
        .assign(rule_2 = np.where(df_results['pct.missing'] > 0, 25, 0))
        .assign(rule_3 = np.where(df_results['min'] == df_results['75%'], 80, 0))
        .assign(rule_4 = np.where(df_results['min'] == df_results['50%'] , 60, 0))
        .assign(rule_4 = lambda X: np.where(X['50%'] == X['75%'] , 0, X['rule_4']))
        .assign(rule_5 = np.where(df_results['min'] < 0, 60, 0))
        .assign(rule_6 = list(map(check_dtype, df_results['type'], df_results['unique'])))
    )
    
    if target:
        df_results.loc[target, ['rule_3', 'rule_4', 'rule_6']] = 0

    df_results = (
        df_results
        .assign(quality_score = lambda X: 1 - ((X['rule_1'] + X['rule_2'] + X['rule_3'] + X['rule_4'] + X['rule_5'] + X['rule_6']) / 350))
    )
    
    cols = ['rule_1', 'rule_2', 'rule_3', 'rule_4', 'rule_5', 'rule_6']

    df_results[cols] = df_results[cols].where(df_results[cols]==0,1)
    
    df_results['pct.missing'] = df_results['pct.missing'].apply(lambda x: '{:.1%}'.format(x))
    
    log = config.get_logger()
    
    log.info(f'rule_1 : More than 50% of the data missing')
    log.info(f'rule_2 : Missing some data')
    log.info(f'rule_3 : 75% of the data is the same and equal to the minimum')
    log.info(f'rule_4 : 50% of the data is the same and equal to the minimum')
    log.info(f'rule_5 : Has negative values')
    log.info(f'rule_6 : Possible wrong data type')
    
    df_results = df_results.sort_values('quality_score', ascending=True)
    
    return df_results


class binary_eda_plot:
    
    """
    
    Visualizing data for explatory analysis.
    
    Parameters
    ----------
        
    data : pandas.DataFrame
        Data set for explatory analysis.
        
    plot_columns : dict (default=None) 
        Columns to visualize.
        
    log_columns : list (default=[None]) 
        Columns to use log scale.

    exclude_cols : list (default=[None]) 
        Columns to not to plot.
        
    columns : int (default=6)
        Number of columns in the matplotlib subplots.
    
    target_palette : dict (default = {1:'red', 0:'deepskyblue'}) 
        Palette for the labels.
    
    Returns
    -------
    
    plot: matplotlib plot
        
    """
    
    def __init__(self, data, 
                 target='target', 
                 plot_columns=None, 
                 log_columns=[None], 
                 exclude_cols=[None], 
                 columns=6, 
                 target_palette={1:'red', 0:'deepskyblue'},
                 bin_numeric=True,
                 sort_by_index=False
                ):
        
        self.data = data
        
        self.target = target
        
        self.exclude_cols = exclude_cols

        self.sort_by_index = sort_by_index

        if not plot_columns:
            
            self.plot_columns = {'target': [], 'discrete' : [], 'numeric': []}
            
            for col in self.data.columns:
            
                if col==self.target:
                    self.plot_columns['target']= col

                elif col not in self.exclude_cols:

                    if len(self.data[col].unique()) < 50 :
                        self.plot_columns['discrete'].append(col)

                    elif self.data[col].dtype in ('float', 'int'):
                        self.plot_columns['numeric'].append(col)
        else:
            self.plot_columns = plot_columns
        
        self.log_columns = log_columns
        
        # self.columns = columns
        
        self.target = target
        
        self.bin_numeric = bin_numeric
        
        self.target_palette = target_palette

        
        self.plot_vars = [self.plot_columns.get('target') ] + self.plot_columns.get('discrete') + self.plot_columns.get('numeric') 
        
        self.plot_vars = sorted(set(self.plot_vars), key=self.plot_vars.index)
        
        # double number of discrete variables for split between labels
        
        self.numb_vars = len(self.plot_vars) 
        self.columns = self.numb_vars
        
        # self.rows = math.ceil(self.numb_vars/self.columns)
        self.rows = 1
        
        self.length = self.rows * 6
        
        self.width = self.columns * 6
        
        
    def label_bar(self, axes, size):
        
        for ax in axes.patches:
            
            if ax.get_height()>0:
                
                value = '{:.0%}'.format(ax.get_height() / size)

                axes.text(ax.get_x() + (ax.get_width() * 0.5 )
                         ,ax.get_height() + (ax.get_height() * 0.025)
                         ,value
                         ,ha='center' 
                         ,va='center'
                         ,fontsize=10
                        )

                
    def format_ax(ax, colors = ['black', 'black', 'black', 'None']):
        
        ax.spines['left'].set_color(colors[0])
        
        ax.spines['left'].set_linewidth(1.5)
        
        ax.spines['bottom'].set_color(colors[1])
        
        ax.spines['bottom'].set_linewidth(1.5)
        
        ax.spines['right'].set_color(colors[1])
        
        ax.spines['right'].set_linewidth(1.5)
        
        ax.spines['top'].set_color(None)
  


    def plot_target(self, ax):
        
        sns.countplot(x=self.target, 
                      data=self.data, 
                      hue=self.target,
                      dodge=False, 
                      palette=self.target_palette, 
                      ax=ax
                     )
        
        plt.ticklabel_format(style='plain', axis='y')
        
        plt.title(self.target, fontsize=14, fontname='georgia', fontweight='bold')
        
        plt.xlabel('')
        
        self.label_bar(ax, len(self.data))
        
        plt.legend(loc='upper right')
        
    def create_bins(self, col):
        
        data = self.data[[col, self.target]]
        
#         # lower limit
#         bins = [data[col].min()]

#         qtiles = list(data[col].quantile([0.1, 0.25, 0.5, 0.75, 0.9]).values)

#         bins.extend(qtiles)
        
#         # upper limit
#         bins.append(data[col].max())
        
        bins = list(data[col].quantile([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]).values)

#         pc_labels = [' (0%)', ' (10%()', ' (25%)', ' ( 50%)', ' (75%)', ' (90%)', ' (100%)']
        
        pc_labels = [f' ({i}%)' for i in range(0,110,10)]

        dic_Labels = dict(list(zip(bins, pc_labels)))

        labels = []

        uniq_bins = sorted(list(set(bins)))

        for i in range(1, len(uniq_bins)):
            
            bin_ = uniq_bins[i]
            
            label = 'upto_' + str(uniq_bins[i]) + dic_Labels[bin_]
            
            labels.append(label)
            
        bin_column = col + "_bin"
            
        data[bin_column] = pd.cut(data[col], bins=bins, include_lowest=True, duplicates='drop', labels=labels)
        
        return data
    

    def gen_plot_data(self, col, data=pd.DataFrame()):
        
        if len(data) == 0:
            data=self.data
        
        a = data[col].value_counts()
        if self.sort_by_index:
            a = a.sort_index()
        a.name = 'total'
        
        b = data[data[self.target]==1][col].value_counts()
        b.name = 'goods'
        
        c = data[data[self.target]==1][col].value_counts(normalize=True)
        c.name = 'distr_goods'
        
        d = data[data[self.target]==0][col].value_counts()
        d.name = 'bads'
        
        e = data[data[self.target]==0][col].value_counts(normalize=True)
        e.name = 'distr_bads'
        
        f = data.groupby(col)[self.target].mean()
        f.name = 'target_rate'

        g = a/len(data)
        g.name = 'attr_rate'

        results = pd.concat([a, b, c, d, e, f, g], axis=1)

        results.index = results.index.map(str)

        return results
    
        
    def plot_discrete(self, results, col, ax):
        
        ax.bar(results.index, results['goods'], color='#ff2e63')
        
        ax.bar(results.index, results['bads'], bottom=results['goods'], color='deepskyblue')
        
#         self.format_ax(ax)
        
        ax2 = ax.twinx()
        
        ax2.plot(results['distr_goods'], marker='o', color='#0000FF', label='event_rate')
        
        ax2.plot(results['distr_bads'], marker='o', color='black', label='non-event_rate')
        
        ax2.plot(results['target_rate'], marker='o', color='red', label='target_rate')

        # ax2.plot(results['attr_rate'], marker='o', color='#00FF00', label='attr_rate')
        
        ax2.set_yticklabels(['{:.0%}'.format(x) for x in ax2.get_yticks()])
        
        ax2.grid(False)
        
        ax.tick_params(axis='x', rotation=90)
        
        plt.title(col, fontsize=14, fontname='georgia', fontweight='bold')
        
        ax2.legend()
        
        plt.xlabel('')
        
        
    def plot_numeric(self, col, ax):
        
        data = remove_outliers(self.data, [col]).percentile_method(threshold=0.95)
        
        sns.kdeplot(
            data=data,
            x=col,
            hue='target',
            log_scale=True if col in self.log_columns else False,
            fill=True,
            palette=self.target_palette,
            ax=ax,
            legend = False
        )  
        
        plt.title(col, fontsize=14, fontname='georgia', fontweight='bold')
        
        plt.xlabel('')
        
        
    def get_plots(self):  

        fig2 = plt.figure(figsize=(self.width, self.length), constrained_layout=True)

        j = 0

        for n, col in enumerate(self.plot_vars):

            if self.plot_columns.get('target') == col:

                j += 1

                axes = plt.subplot(self.rows, self.columns, j)
                
                self.plot_target(axes)

            elif col in self.plot_columns.get('discrete'):

                j += 1

                ax1 = plt.subplot(self.rows, self.columns, j)
                
                results = self.gen_plot_data(col=col, data=pd.DataFrame())
                
                self.plot_discrete(results=results, col=col, ax=ax1)

            elif col in self.plot_columns.get('numeric'):

                j += 1

                ax2 = plt.subplot(self.rows, self.columns, j)
                
                if self.bin_numeric:
                    
                    data = self.create_bins(col)
                    
                    col_name = col + '_bin'
                    
                    results = self.gen_plot_data(col=col_name, data=data)
                    
                    self.plot_discrete(results=results, col=col_name, ax=ax2)
                
                else:
                    
                    self.plot_numeric(col, ax2)

        plt.tight_layout()





import pandas as pd 
# import snowflake.connector

from snowflake.sqlalchemy import URL
from sqlalchemy import create_engine
from cryptography.fernet import Fernet


from ml_processor.configuration import config
import time
import os
import json



class snowflake_processor:

    """

    Performing ETL tasks such as connecting to snowflake and retrieving data from snowflake.
    
    Parameters
    ----------

    credentials : dick (default=None) 
        Dictionary with connection credentials e.g
        >>> {'username':'myUserName', 'password':'******', 'account':'snowfalkeAccount', 'warehouse':'warehouseName', 'database':'dataName'}
        
    credential_path : string  (default=None) 
        Path to credentials stored in an enrypted file. This should only be provided if credentials have not been provided. 
        The contents of the file should be a dictionary encrypted using Fernet.

    key_path : string (default=None)  
        Path to Fernet key for decrypting the file provided using the credential_path..
        
    """

    def __init__(self, 
        credentials = None, 
        credential_path = None,
        key_path = None,
        ):
        
        if credentials:
            self.credentials = credentials
        else:
            if not credential_path:
                credential_path = os.path.join(os.environ["HOME"], 'Desktop/package_dev/credentials/encrypted_credentials.csv')

            if not key_path:
                key_path = os.path.join(os.environ["HOME"], 'Desktop/package_dev/credentials/sf_key.key')

            with open(key_path, 'rb') as keyfile:
                key = keyfile.read()
            
            with open(credential_path, 'rb') as credfile:
                creds = credfile.read()

            fernet = Fernet(key)

            creds = fernet.decrypt(creds).decode()

            self.credentials = json.loads(creds)

        self.logger = config.get_logger()

    def connect(self):
        
        """
        
        Create connection to snowflake.

        Parameters
        ----------

        None
        
        Returns
        -------
            
        object
            connection to snowflake.

        """
##### connection using sqlalchemy
        try:
            engine = create_engine(
                URL(
                    account = self.credentials.get('account'),
                    user = self.credentials.get('username'),
                    password = self.credentials.get('password'),
                    database = self.credentials.get('database'),
                    warehouse = self.credentials.get('warehouse'),
                    schema = self.credentials.get('schema'),
                )
            )
            self.logger.info(f'Connection to {self.credentials.get("account")} successful')
        except:
            self.logger.error('Exception occured', exc_info=True)
        else:
            return engine

##### connection using snowflake.connector
        # try:
        #     engine = snowflake.connector.connect(
        #         user = self.credentials.get('username'),
        #         password = self.credentials.get('password'),
        #         account = self.credentials.get('account'),
        #         database = self.credentials.get('database'),
        #         warehouse = self.credentials.get('warehouse'),
        #         )

        #     self.logger.info(f'Connection to {self.credentials.get("account")} successful')
        # except:
        #     self.logger.error('Exception occured', exc_info=True)
        # else:
        #     return engine

        
    def pandas_from_sql(self, sql, conn=None, chunksize=None):
        
        """

        Extracting data from snowflake into pandas dataframe.
        
        Parameters
        ----------

        sql : string 
            Sql statememt for extracting data
            
        conn : object (default=None) 
            Connection engine to snowflake.
            
        chunksize : int (default=None)
            Number of rows to extract from snowflake per iteration if extracting in chunks.
        
        
        Returns
        -------

        Pandas.DataFrame
            Data extracted fro snowflake.

        """
        
        if not conn:
            conn = self.connect()
            
        start = int(time.time())
            
        if not chunksize:
            
            df = pd.read_sql_query(sql, conn)
        else:
            
            df = pd.DataFrame()
            
            rows = 0
            
            for chunk in pd.read_sql_query(sql, conn, chunksize=chunksize):
                
                df = pd.concat([df, chunk])
                
                rows += chunk.shape[0]
                
        end = int(time.time())
        
        self.logger.info(f'Number of rows extracted: {df.shape[0]}')
        
        self.logger.info(f'Runtime for data extraction : {int(end-start)} seconds')
        
        return pd.DataFrame(df)



from setuptools import setup
import os



path = os.path.join(os.environ.get('HOME'), 'Desktop/package_dev/ml/README.rst')


with open(path, "r") as fh:
    long_description = fh.read()

setup(
    # name that will be imported, can be different from code file name
    name='ml_processor',

    version='0.5.4',

    description='Includes functions for performing econometrics tasks',

    # code file name without file extension
    # py_modules=['configuration', 'eda_analysis', 'encoders', 'jsonSerializer', 'model_training' 'outliers' 'snowflake_processor'],

    packages=['ml_processor'],

    # directory in which code file is stored
    package_dir={'':'src'},

    long_description=long_description,

    # long_description_content_type="text/markdown",

    long_description_content_type="text/x-rst",

    author="Geofrey Wanyama",

    author_email="wanyamag17@gmail.com",

    url="https://github.com/G-Geofrey/package_dev/tree/master/ml",

    zip_safe=False,

    classifers=[ 
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7", 
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
        ],

     install_requires=[
        'pandas',
        'numpy>=1.16',
        'matplotlib',
        'seaborn',
        #'sklearn',
        'optbinning',
        'snowflake-connector-python',
        'snowflake.sqlalchemy',
        'sqlalchemy',
        'cryptography>=3.4.8',
        'joblib',
        'xgboost',
        'statsmodels',
        'hyperopt>=0.2.7',
        'scipy',
        'shap',
        'lightgbm'
    ]


    )



import pandas as pd 
import pandas
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import seaborn as sns

import warnings
import os
import json
import joblib
import time
import copy

import shap

from datetime import datetime
from collections import defaultdict

import xgboost as xgb
from sklearn.ensemble import RandomForestClassifier
from lightgbm import LGBMClassifier

from sklearn.model_selection import train_test_split, StratifiedKFold, cross_val_score
from hyperopt import tpe, STATUS_OK, Trials, hp, fmin, space_eval
from sklearn import metrics
from scipy.interpolate import make_interp_spline

from ml_processor.configuration import config
from ml_processor.jsonSerializer import NpEncoder

sns.set_style('whitegrid')
warnings.filterwarnings('ignore')


# <<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
class model_training:
    
    """

    Performing most machine learning tasks including hyperparameter tuning, model fitting and diagnostic checks.
    
    Parameters
    ----------
        
    df : pandas.Dataframe 
        Dataset with features and labels.

    features : list or array-like 
        Variable names (features) for fitting the model.

    target : string (default = "target")
        Name of column with labels (dependent variable).

    params_prop : float (default = 0.25) 
        Proportion of data set to use for hyperparameter tuning.

    test_size : float (default = 0.33) 
        Proportion of data to use as the test set.

    hyperparams : dictionary (default = None) 
        Predefined hyperparameters and their values.Specified if hyperparameter tunning is not necessary.
    
    method : string (default = "hyperopt", options = ["hyperopt", "random_search", "grid_search"])
        Hyperparameter optimization method

    classifier : string (default = "xgboost", options = ["xgboost", "random_forest", "lightgbm"])
        Classification algorithm to optimize

    eval_metric : string (default = "recall")
        Performance metric to optimise.

    search_space : dic (default = None)
        Range of values for the different parameters to tune.
    
    """

    
    def __init__(self, df, features, target="target", params_prop=0.25, test_size=0.33, hyperparams=None, 
           method="hyperopt", classifier="xgboost", eval_metric="recall", search_space=None
           ):
        
        self.data = df
        
        self.features = features
        
        self.target = target
        
        self.hyperparams = hyperparams

        self.method = method
        
        self.classifier = classifier
        
        self.eval_metric = eval_metric
        
        self.search_space = search_space
        
        self.cwd = os.getcwd()
        
        self.log = config.get_logger(file=f"{self.classifier}.log", folder=f"{self.classifier}_model_artefacts")
        
        if not os.path.isdir(f'{self.classifier}_model_artefacts'):
            os.mkdir(f'{self.classifier}_model_artefacts')

        self.path = os.path.join(f'{self.classifier}_model_artefacts', f'{self.classifier}_Model_results.json')
        
        try:
            with open(self.path) as file:
                self.model_artefacts = json.load(file)
        except:
            self.model_artefacts = {}

        self.model_name = f'{self.classifier}_' + str(datetime.now().strftime('%Y%m%d%H%M%S'))
        
        self.log.info(f'Model job name: {self.model_name}')
        
        self.model_artefacts[self.model_name] = {}
        
        self.params_prop = params_prop
        self.model_artefacts[self.model_name]['tunning_data_size'] = self.params_prop
        
        self.test_size = test_size
        self.model_artefacts[self.model_name]['test_data_size'] = self.test_size

        self.X_train = pd.DataFrame()

        self.df_features = pd.DataFrame()


    def fit_model(self):
        
        """

        Fit model 
        
        Parameters
        ---------- 
        
        None
        
        Returns
        -------
        
        model : object
            Model object
        
        """
        
        if self.X_train.empty:
            self._split_data()
        
        if not self.hyperparams:
            self._parameter_tunning()

        best_params = self.hyperparams
        
        self.log.info('Model fitting initialized...')

        if self.classifier == "xgboost":
            model = xgb.XGBClassifier(**best_params)
        elif self.classifier == "random_forest":
            model = RandomForestClassifier(**best_params)
        elif self.classifier == "lightgbm":
            model = LGBMClassifier(**best_params)
        
        self.log.info('Model fitting started...')

        model.fit(self.X_train[self.features], self.y_train)
        
        # end = int(time.time())
        
        self.log.info('Model fitting completed')
        
        # self.log.info(f'Runtime for fitting the model : {int(end-start)} seconds')
        
        self.model = model

        self._get_metrics()

        # saving all model artefacts
        json_object = json.dumps(self.model_artefacts, cls=NpEncoder, indent=4)
        with open(self.path, 'w') as outfile:
            outfile.write(json_object)
        self.log.info(f'Model artifacts saved @:{self.path}')

        path = os.path.join(os.getcwd(), f'{self.classifier}_model_artefacts', str(self.model_name) + '.sav')

        joblib.dump(model, path)

        self.log.info(f'Model saved @:{path}')

    def model_results():

        """

        Get model results from fitted model.

        Parameters
        ---------- 
        
        None
        
        Returns
        -------
        
        object:
            Model object.
        
        """

        return self.model


    def model_evaluation(self, model = None, features = None, target = "target", train_set = pd.DataFrame(), 
        validation_set = pd.DataFrame(), test_set = pd.DataFrame(), main_set = "test", loc_path = None
        ):

        """

        Generate model diagnostics plots

        Parameters
        ----------

        model : object
            Model to evaluate

        features : list or array-like (default=None)
            Features used for fitting the model

        target : string (default="target")
            COlumn with labels

        train_set : pandas.Dataframe (default=pandas.DataFrame())
            Training set to evaluate the model on. If not set, the model is evlauted on the remaining sets (validation_set, test_set). 
            Atleast one of the sets (training_set, validation_set, test_set) has to be provided. 

        validation_set : pandas.Dataframe (default=pandas.DataFrame())
            Validation set to evaluate the model on. If not set, the model is evlauted on the remaining sets (train_set, test_set). 
            Atleast one of the sets (training_set, validation_set, test_set) has to be provided. 

        test_set : pandas.Dataframe (default=pandas.DataFrame())
            Testing set to evaluate the model on. If not set, the model is evlauted on the remaining sets (train_set, validation_set). 
            Atleast one of the sets (training_set, validation_set, test_set) has to be provided.

        main_set : string (default="test", options=["test", "validation", "train"])
            Dataset to consider as the actual testing set for the model. Some model performance metrices are ony evaluated on the main set.
            The main set must be part of the data sets (training_set, validation_set, test_set) provided.

        loc_path : string (default=None):
            Directory (folder) to save the plots.

        """

        if not model:
            model = self.model 
        
        if  not features:
            features = self.features
        
        if train_set.empty:
            train_set = pd.concat([self.X_train, self.y_train], axis=1)
        
        if test_set.empty:
            test_set = pd.concat([self.X_test, self.y_test], axis=1)
        
        if not loc_path:
            loc_path = os.path.join(f"{self.classifier}_model_artefacts", "model_performance_plot.png")

        model_perf = plot_model_perf(
            model = model,
            features = features,
            target = target,
            train_set = train_set,
            validation_set = validation_set,
            test_set = test_set,
            main_set = main_set,
            loc_path = loc_path,
            )

        return model_perf.make_plots()


    def shap_summary_plot(self, model=None, eval_set=pd.DataFrame(), save_plot=True, save_path=None, plot_size=(15,20) ):

        """

        Generate shap summary plot for fitted model

        model : object 
            Fitted model whose shap summary plot is to be generated

        eval_set : pandas.DataFrame (default=pandas.DataFrame())
            Data set to use for model evaluation

        save_plot : boolean (default=True)  
            Whether to save plot or not

        save_path : string (default = None)
            Path to save plot to.

        plot_size : "auto", float, (float, float) (default), or None.
            What size to make the plot.

        """

        if not model:
            model = self.model

        if eval_set.empty:
            eval_set = self.X_train

        if save_plot:
            if not save_path:
                save_path = os.path.join(f"{self.classifier}_model_artefacts", "shap_summary_plot.png")

        model_interp_shap(model=model, eval_set=eval_set, save_path=save_path, plot_size=plot_size )


    def _parameter_tunning(self):
        
        """

        Hyperparameter tuning using hyperopt method.

        Parameters
        ----------
        
        None
        
        Returns
        -------
        
        None
        
        """

        self._split_tunning_data()
        

        self.hyperparams = hyper_parameter_tunning(
            features = self.X_train_param[self.features], 
            labels = self.y_train_param, 
            method = self.method,
            classifier = self.classifier,
            eval_metric = self.eval_metric,
            search_space = self.search_space,
            )
        
        self.log.info('Hyperparameter tuning completed')
        
        # self.log.info(f'Runtime for Hyperparameter tuning : {int(end-start)} seconds')

        # self.log.info(f'Best parameters: {self.hyperparams}')
        
        self.model_artefacts[self.model_name]['best_params'] = self.hyperparams



    def _get_metrics(self):
        
        """

        Function for generating feature model performance metrices for fitted model 

        Parameters
        ----------
        
        None
        
        Returns
        -------
        
        None
        
        """
        
        if self.df_features.empty:
            self._feature_imp()

        metrices = get_metrics(self.model, self.X_test, self.y_test)

        predictions = get_predictions(self.model, self.X_test, self.y_test)

        recall_precision_arrays = get_prediction_arrays(self.model, self.X_test, self.y_test)
        
        self.pred_class = predictions["pred_class"]
        self.log.info('Predicted labels generated (test)')
        

        self.pred_prob = predictions["pred_prob"]
        self.log.info('Predicted probabilities generated (test)')

        self.confMatrix = metrices["confusion_matrix"]
        self.log.info('Confusion matrix generated (test)')

        self.area_roc = metrices["area_roc"]
        self.model_artefacts[self.model_name]['area_roc'] = float(self.area_roc)
        self.log.info(f"AUC (test): {self.area_roc:.0%}")

        metrices_ = ["precision", "recall", "fscore", "accuracy", "false_positive_rate"]
        for metric in metrices_:
            metric_value = float(metrices[metric])
            self.model_artefacts[self.model_name][metric] = metric_value
            self.log.info(f"{metric} (test): {metric_value:.0%}")

        self.precision_curve, self.recall_curve,  = recall_precision_arrays["precision_curve"], recall_precision_arrays["recall_curve"] 
        self.log.info('Precision and Recall values for the precision recall curve created')


        self.trP, self.trN = recall_precision_arrays["trP"], recall_precision_arrays["trN"]
        self.log.info('True positive and negativevalues for the ROC curve created')


    def _feature_imp(self):
        
        """

        Function for generating feature importance for fitted model 
        
        Parameters
        ----------
        
        None
        
        Returns
        -------
        
        None
        
        """

        try:

            feature_importance = get_variableImportance(self.model, self.features) 
            
            self.log.info('Dataframe with feature importance generated')
            
            self.df_features = pd.DataFrame(feature_importance)
            
            self.model_artefacts[self.model_name]['feature_importance'] = {
                'feature' : list(self.features), 
                'importance' : list(self.model.feature_importances_)
            }

        except:
            self.log.warning(f"!!! Model object not found, fit model first and proceed")
            

    def _split_data(self): 
        
        """

        Split data into training and test sets

        Parameters
        ----------
        
        None
        
        Returns
        -------
        
        None
        
        """

        self.X_train, self.X_test, self.y_train, self.y_test = split_data(data=self.data, target=self.target, test_size=self.test_size)
        
        self.log.info('Splitting data into training and testing sets completed')
        
        self.log.info(f'Training data set:{self.X_train.shape[0]} rows')
        
        self.log.info(f'Testing data set:{self.X_test.shape[0]} rows')


    def _reduce_data(self):
        
        """

        Generate a fraction of the data to use for hyperparameter tuning data 

        Parameters
        ----------
        
        None
        
        Returns
        -------
        
        pandas.DataFrame:
            Dataset for hyperparameter tuning
        
        """
        
        tunning_data_size = int(self.data.shape[0] * self.params_prop)
        
        self.log.info('Hyper parameter tuning data set created')
    
        self.log.info(f'Hyper parameter tuning data set:{tunning_data_size} rows')
        
        return self.data.sample(tunning_data_size)


    def _split_tunning_data(self):
        
        """

        Split data into training and test sets

        Parameters
        ----------
        
        None
        
        Returns
        -------
        
        None
        
        """
        
        tunning_data = self._reduce_data()

        self.X_train_param, self.X_test_param, self.y_train_param, self.y_test_param = split_data(data=tunning_data, target=self.target, test_size=self.test_size)
        
        self.log.info('Splitting hyperparameter tuning data into training and testing sets completed')
        
        self.log.info(f'Hyperparameter tuning training data set:{self.X_train_param.shape[0]} rows')
        
        self.log.info(f'Hyperparameter tuning testing data set:{self.X_test_param.shape[0]} rows')



# <<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
class plot_model_perf:

    """

    Generate model diagnostics plots

    Parameters
    ----------

    model : object
        Model to evaluate

    features : list or array-like (default=None)
        Features used for fitting the model

    target : string (default="target")
        COlumn with labels

    train_set : pandas.Dataframe (default=pandas.DataFrame())
        Training set to evaluate the model on. If not set, the model is evlauted on the remaining sets (validation_set, test_set). 
        Atleast one of the sets (training_set, validation_set, test_set) has to be provided. 

    validation_set : pandas.Dataframe (default=pandas.DataFrame())
        Validation set to evaluate the model on. If not set, the model is evlauted on the remaining sets (train_set, test_set). 
        Atleast one of the sets (training_set, validation_set, test_set) has to be provided. 

    test_set : pandas.Dataframe (default=pandas.DataFrame())
        Testing set to evaluate the model on. If not set, the model is evlauted on the remaining sets (train_set, validation_set). 
        Atleast one of the sets (training_set, validation_set, test_set) has to be provided.

    main_set : string (default="test", options=["test", "validation", "train"])
        Dataset to consider as the actual testing set for the model. Some model performance metrices are ony evaluated on the main set.
        The main set must be part of the data sets (training_set, validation_set, test_set) provided.

    loc_path : string (default=None):
        Directory (folder) to save the plots.

    """

    def __init__(self, model, features, target="target", train_set=pd.DataFrame(), 
        validation_set=pd.DataFrame(), test_set=pd.DataFrame(), main_set="test", loc_path=None
        ):


        self.model = model 
        self.features = features
        self.target = target
        self.train_set = train_set
        self.validation_set = validation_set
        self.test_set = test_set
        self.main_set = main_set

        if loc_path:
            self.loc_path = loc_path
        else:
            self.loc_path = os.path.join(os.getcwd(), "model_performance_plot.png")

        self.columnNames = {}

        self.eval_sets = ["train", "validation", "test"]
        self.eval_sets_dic = {}

        # self.main_data_set  = eval(f"self.{self.main_set}_set")

        for eval_set in self.eval_sets:

            data  = eval(f"self.{eval_set}_set")

            if isinstance(data, pandas.core.frame.DataFrame) and not data.empty:

                columnName = f"{eval_set}(size={len(data):,})"
                self.columnNames[eval_set] = columnName

                self.eval_sets_dic[eval_set] = {}

                predictions = get_predictions(self.model, data[self.features], data[self.target])
                self.eval_sets_dic[eval_set]["predictions"] = predictions

                _metrices = get_metrics(model, data[self.features], data[self.target])
                self.eval_sets_dic[eval_set]["_metrices"] = _metrices

                pred_array = get_prediction_arrays(model, data[self.features], data[self.target])
                self.eval_sets_dic[eval_set]["pred_array"] = pred_array

        self.df_features = get_variableImportance(self.model, self.features, returnType="df")

    def get_metric_df(self):

        """

        Generate dataframe with model performance metrics

        Parameters
        ----------
        
        None
        
        Returns
        -------
        
        df_metrices : pandas.DataFrame
            Metrics data
        
        """

        df_metrices = pd.DataFrame()

        for i, eval_set in enumerate(self.eval_sets_dic):
            metric = self.eval_sets_dic[eval_set]["_metrices"].copy()
            del metric["confusion_matrix"]
            metric = pd.Series(metric)
            metric.name = eval_set
            df_metrices = pd.concat([df_metrices, metric], axis=1)
        
        df_metrices = df_metrices.rename(columns = self.columnNames)

        return df_metrices

    def make_plots(self, display_metrics=True):

        """

        Generate model performance plots

        Parameters
        ----------
        
        display_metrics : boolean (default=True)
            Whether to display metrics table or not.
        
        Returns
        -------
        
        matplotlib plot 
        
        """

        if display_metrics:
            display(self.get_metric_df())

        sns.set_style('whitegrid')
        colors = {"train":"black", "validation":"orange", "test":"deepskyblue"}

        if self.df_features.empty:
            width = 6
            rows = 1
        else:
            width = 12
            rows = 2
        
        fig = plt.figure(figsize=(25, width), constrained_layout=True)

        spec2 = gridspec.GridSpec(ncols=5, nrows=rows, figure=fig)
        
        fig_ax1 = fig.add_subplot(spec2[0,0])
        fig_ax2 = fig.add_subplot(spec2[0,1])
        fig_ax3 = fig.add_subplot(spec2[0,2])
        fig_ax4 = fig.add_subplot(spec2[0,3])
        fig_ax5 = fig.add_subplot(spec2[0,4])

        # confusion matrix
        confMatrix = self.eval_sets_dic[self.main_set]["_metrices"].get("confusion_matrix")
        _plot_comf_matrix(confMatrix, ax=fig_ax1, dataset=self.main_set)

        # predictions
        pred_prob = self.eval_sets_dic[self.main_set]["predictions"].get("pred_prob")
        pred_class = self.eval_sets_dic[self.main_set]["predictions"].get("pred_class")
        y_test = eval(f"self.{self.main_set}_set")[self.target]
        _plot_prediction(y_test, pred_class, pred_prob, ax=fig_ax4, dataset=self.main_set)

        for i, eval_set in enumerate(self.eval_sets_dic):
            trP = self.eval_sets_dic[eval_set]["pred_array"].get("trP")
            trN = self.eval_sets_dic[eval_set]["pred_array"].get("trN")

            area_roc = self.eval_sets_dic[eval_set]["_metrices"].get("area_roc")

            recall_curve = self.eval_sets_dic[eval_set]["pred_array"].get("recall_curve")
            precision_curve = self.eval_sets_dic[eval_set]["pred_array"].get("precision_curve")
            # AP = self.eval_sets_dic[eval_set]["pred_array"].get("AP")

            color = colors.get(eval_set)

            _plot_roc_curve(trP, trN, area_roc, ax=fig_ax2, color=color, label=eval_set)
            _plot_rp_curve(recall_curve, precision_curve, ax=fig_ax3, color=color, label=eval_set)

        _plot_lift_curve(y_test, pred_prob, ax=fig_ax5, dataset=self.main_set)

        if not self.df_features.empty:
            
            fig_ax6 = fig.add_subplot(spec2[1,:])
            _plot_feature_imp(self.df_features, ax=fig_ax6)

        plt.subplots_adjust(hspace=0.3, wspace=0.3)

        plt.savefig(self.loc_path)


# <<<<<<<<<<<<<<<<<<<<<<<<<<<<--------------------------------------------

def split_data(data, target="target", test_size=0.33): 
    
    """

    Split data into training and test sets

    Parameters
    ----------
    
    df : pandas.Dataframe 
        Full dataset to split into training and testing set. Includes both features and labels.

    target : string (default = "target")
        Name of column with labels (dependent variable).

    test_size : float (default=0.33) 
        Proportion of data to use as the test set.

    Returns
    -------
    
    X_train, X_test, y_train, y_test : pandas.DataFrames
        Where ;
            X_train : training features data
            X_test  : testing features data
            y_train : training labels data
            y_test  : testing labels data
    
    """
    
    X = data.drop(target, axis=1)
    
    y = data[target]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size)

    return X_train, X_test, y_train, y_test



def model_interp_shap(model, eval_set, save_path=None, plot_size=(10,10)):

    """

    Generate shap summary plot for fitted model

    model : object 
        Fitted model whose shap summary plot is to be generated

    eval_set : pandas.DataFrame (default=pandas.DataFrame())
        Data set to use for model evaluation

    save_path : string (default = None)
        Path to save plot to.

    plot_size : "auto", float, (float, float) (default), or None.
        What size to make the plot.

    """

    features = eval_set.columns
    num_cols = len(features)
    
    if isinstance(model, xgb.sklearn.XGBClassifier):
        
        explainer = shap.TreeExplainer(model)
        
        shap_values = explainer.shap_values(eval_set)

        shap.summary_plot(shap_values, eval_set, max_display=num_cols, show=False, plot_size=plot_size)

        if save_path:
            plt.savefig(save_path, bbox_inches='tight',dpi=100)

        plt.show()
    
    elif isinstance(model, LGBMClassifier) or isinstance(model, RandomForestClassifier):
        
        explainer = shap.TreeExplainer(model)
        
        shap_values_ = explainer(eval_set)
        
        shap_values = copy.deepcopy(shap_values_)
        
        shap_values.values = shap_values.values[:,:,1]
        
        shap_values.base_values = shap_values.base_values[:,1]
        
        shap_plot = shap.plots.beeswarm(shap_values, max_display=num_cols, show=False, plot_size=plot_size)

        if save_path:
            plt.savefig(save_path, bbox_inches='tight',dpi=100)

        plt.show()



def hyper_parameter_tunning(features, labels, method="hyperopt", classifier="xgboost", 
        eval_metric="recall", search_space=None):
    
    """

    Hyperparameter tuning for specified algorithm

    Parameters
    ----------
    
    features : pandas.Dataframe 
        Dataset with features

    labels : pandas.Dataframe 
        Dataset with features

    # method : string (default = "hyperopt", options = ["hyperopt", "random_search", "grid_search"])
    #     Hyperparameter optimization method --- in future versions

    classifier : string (default = "xgboost", options = ["xgboost", "random_forest", "lightgbm"])
        Classification algorithm to optimize

    eval_metric : string (default = "recall")
        Performance metric to optimise.

    search_space : dic (default = None)
        Range of values for the different parameters to tune.
    
    Returns
    -------
    
    dic :
        Hyperparameter values
    
    """

    if not search_space:
        space = search_space_generator(classifier)
    else :
        space = search_space
    
    kfold = StratifiedKFold(n_splits=3, shuffle=True, random_state=42)
    
    def objective_func(params):
        
            if classifier == "xgboost":
                model = xgb.XGBClassifier(**params)
            elif classifier == "random_forest":
                model = RandomForestClassifier(**params)
            elif classifier == "lightgbm":
                model = LGBMClassifier(**params)

            scores = cross_val_score(
                model,
                features, 
                labels,
                cv=kfold, 
                scoring=eval_metric, 
                n_jobs=-1
            )

            best_score = np.mean(scores)

            loss = - best_score

            return {'loss':loss, 'params':params, 'status':STATUS_OK}
    
    print('Trials initialized...')
    
    trials = Trials()

    param_search = fmin(fn = objective_func, space=space, algo=tpe.suggest, max_evals=48, trials=trials)

    hyperparams = space_eval(space, param_search)

    return hyperparams



def get_predictions(model, X_test, y_test, threshold=0.5):

    """

    Generate probability and class predictions from fitted model

    Parameters
    ----------

    model : object
        Model to use for making predictons

    X_test : pandas.DataFrame
        Data set to make predictions on; should include only features used in fitting the model.

    y_test : list or array-like
        Test labels

    threshold : float (default=0.5)
        Probability cut-off value for class allocation 
        e.g threshold=0.5 means all samples where the predicted probability is atleast 0.5 are allocated the positive label.

    Returns
    -------

    predictions : dic
        Predicted probabilities and classes 
            • pred_prob - predicted probabilities
            • pred_class - predicted classes

    """

    if isinstance(model, xgb.core.Booster):
        
        data = xgb.DMatrix(X_test, y_test)
        
        pred_prob = model.predict(data)
    
    elif not hasattr(model, "predict_proba") and hasattr(model, "predict"):
        
        pred_prob = model.predict(X_test)
    
    elif hasattr(model, "predict_proba") and hasattr(model, "predict"):
        
        pred_prob = model.predict_proba(X_test)[:,1]

    pred_class = pred_prob>threshold
    
    pred_class = pred_class.astype(int)

    predictions = {"pred_prob": pred_prob, "pred_class": pred_class}

    return predictions



def get_metrics(model, X_test, y_test):

    """

    Generate model performance metrics

    Parameters
    ----------

    model : object
        Model for which to obtain performance metrics

    X_test : pandas.DataFrame
        Data set to make predictions on; should include only features used in fitting the model.

    y_test : list or array-like
        Test labels


    Returns
    -------

    _metrics : dic
        Model performance metrices
            • area_roc - Area under the ROC Curve.
            • accuracy - accuracy
            • precision - precision
            • recall - Recall/sensitivity
            • fscore - F1-score
            • false_positive_rate - False positive rate
            • confusion_matrix - confusion matrix 


    """

    predictions = get_predictions(model, X_test, y_test)
    pred_prob = predictions.get("pred_prob")
    pred_class = predictions.get("pred_class")

    area_roc = metrics.roc_auc_score(y_test, pred_prob)
    
    accuracy = metrics.accuracy_score(y_test, pred_class)
    
    precision, recall, fscore, support = metrics.precision_recall_fscore_support(y_test, pred_class)
    
    confMatrix = metrics.confusion_matrix(y_test, pred_class)
    
    false_positive_rate = confMatrix[0,1]/confMatrix[0,:].sum()

    _metrics = {
        "area_roc": area_roc,
        "accuracy": accuracy,
        "precision": precision[1],
        "recall": recall[1],
        "fscore": fscore[1],
        "false_positive_rate": false_positive_rate,
        "confusion_matrix": confMatrix
    }

    return _metrics



def get_prediction_arrays(model, X_test, y_test):

    """

    Genertae true positive-true negative, precision-recall pairs for different probability thresholds.


    Parameters
    ----------

    model : object
        Model for which to obtain performance metrics

    X_test : pandas.DataFrame
        Data set to make predictions on; should include only features used in fitting the model.

    y_test : list or array-like
        Test labels


    Returns
    -------

    pred_array : dic
        True positive, true negative, precision, recall arrays
            • trP - True positive array
            • trN - True negative  array
            • precision_curve - precision  array
            • recall_curve - recall  array

    """

    predictions = get_predictions(model, X_test, y_test)
    pred_prob = predictions.get("pred_prob")
    pred_class = predictions.get("pred_class")
    
    trP, trN, _ = metrics.roc_curve(y_test, pred_prob)
    
    precision_curve, recall_curve, _ = metrics.precision_recall_curve(y_test, pred_prob)

    pred_array = {
        "trP": trP,
        "trN": trN,
        "precision_curve": precision_curve,
        "recall_curve": recall_curve,
    }
    
    return pred_array



def get_variableImportance(model, features, returnType="dic"):

    """
    Generate variable importance for each variable in fitting a model


    Parameters
    ----------

    model : object
        Model for which to obtain variable importance

    features : list or array-like
        Features used fro fitting model; should be in the same order as in the model fitting data set.

    returnType : string (defaul="dic", options=["dic", "df"])
        Returns dictionary if set to "dic" and pandas.DataFrame if set to "df"

    Returns
    -------

    featureImportance/df_featureImp : dic/pandas.DataFrame
        Variable importance values for each variable

    """

    if isinstance(model, xgb.core.Booster):
        feature_scores = model.get_score(importance_type='gain')
        featureImportance = {"feature" : feature_scores.keys(), 'importance' : feature_scores.values()}
    
    elif hasattr(model, "feature_importances_"):
        featureImportance = {"feature" : features, 'importance' : model.feature_importances_}
    
    else:
        featureImportance = {}

    if returnType == "dic":
        return featureImportance
    elif returnType == "df":
        df_featureImp = pd.DataFrame(featureImportance)

        if not df_featureImp.empty:
            df_featureImp = df_featureImp.sort_values("importance", ascending=False)
        
        return df_featureImp



def search_space_generator(classifier="xgboost"):
    """

    Generates search space for hyperparameter tuning for specified algorithm


    Parameters
    ----------
    
    classifier : string (default = "xgboost)
    
    Returns
    -------
    
    dic:
        Range of hyperparameters to search from
    
        """
    space = {
        "xgboost" : {
            'n_estimators': hp.choice('n_estimators', range(100,1500,100)),
            'learning_rate': hp.choice('learning_rate', np.arange(0.01, 0.5, 0.01)),
            'max_depth': hp.choice('max_depth', range(2, 15, 1)),
            'subsample': hp.choice('subsample', np.arange(0.1, 1.1, 0.1)),
            'colsample_bytree': hp.choice('colsample_bytree', np.arange(0.1, 1.1, 0.1)),
            'gamma': hp.choice('gamma', np.arange(0.1, 0.6, 0.1)),
            'reg_alpha': hp.choice('reg_alpha', [1e-5, 1e-2, 0.1, 1, 10, 100]),
            'reg_lambda': hp.choice('reg_lambda', [1e-5, 1e-2, 0.1, 1, 10, 100]),
            'objective': 'binary:logistic',
            'booster': 'gbtree',
            'importance_type': 'gain',
            'random_state': 801,
            'seed': 0,
            'verbosity': 1,
            'n_jobs': -1,
        },

        "random_forest" : {
            'n_estimators': hp.choice('n_estimators', range(100,1500,100)),
            'max_depth': hp.choice('max_depth', range(2, 15, 1)),
            'criterion': 'gini',
            'min_samples_split': hp.choice('min_samples_split', [2, 5, 10]),
            'max_features' : hp.choice('max_features', ['auto', 'sqrt']),
            'bootstrap': True,
            # 'importance_type': 'gain',
            'random_state': 801,
            'verbose' : 0,
            'n_jobs' : -1,
            
        },

        "lightgbm" : {
            'num_leaves' : hp.choice('num_leaves', range(20, 220, 20)),
            'max_depth': hp.choice('max_depth', range(2, 15, 1)),
            'learning_rate' : hp.choice('learning_rate', np.arange(0.01, 0.5, 0.01)),
            'n_estimators': hp.choice('n_estimators', range(100,1500,100)),
            'subsample' : hp.choice('subsample', np.arange(0.6, 1, 0.05)),
            'min_child_samples' : hp.choice('min_child_samples', range(5,50,10)),
            'boosting_type' : 'gbdt',
            'objective': 'binary',
            'importance_type': 'gain',
            'random_state': 801,
            'seed': 0,
            'n_jobs': -1,
        }
    }

    return space.get(classifier)

# ------------------------------------------->>>>>>>>>>>>>>>>>>


def _plot_comf_matrix(cm, ax, dataset="test"):
        
    sns.heatmap(cm, annot=True, cmap="Blues", fmt="d", cbar=False, annot_kws={"size": 12}, ax=ax)
    
    ax.set_xlabel('Predicted label', fontsize=14)
    
    ax.set_ylabel('True label', fontsize=14)

    ax.set_title(f"Confusion matrix ({dataset})", fontsize=16)
    
    ax.tick_params(axis='both', which='major', labelsize=12)
    
    ax.set_frame_on(False)


def _plot_roc_curve(trP, trN, area_roc, ax, color="orange", label="test"):
        
    ax.plot(trP, trN, color=color, label= f"{label} (AUC ≈ {area_roc:.3f})")
    
    ax.plot([1,0], [1,0], "--", color='darkblue')
    
    ax.set_frame_on(False)
    
    ax.tick_params(axis='both', which='major', labelsize=12)
    
    ax.set_xlabel('False Positive Rate', fontsize=14)
    
    ax.set_ylabel('True Positive Rate', fontsize=14)
    
    ax.set_title('Receiver Operating Characteristic (ROC) Curve', fontsize=16)
    
    # ax.text(0.3, 0.7, 'AUC = {:.3f}'.format(area_roc), fontsize=12)

    ax.legend(frameon=False)


def _plot_rp_curve(rcl, prc, ax, color="orange", label="test"):
        
    ax.plot(rcl, prc, color=color, label= f"{label}")
    
    ax.set_frame_on(False)
    
    ax.tick_params(axis='both', which='major', labelsize=12)
    
    ax.set_xlabel('recall', fontsize=14)
    
    ax.set_ylabel('precision', fontsize=14)
    
    ax.set_title('Precision recall Curve', fontsize=16)

    ax.legend(frameon=False)


def _plot_prediction(y_test, pred_class, pred_prob, ax, dataset="test"):
    
    palette = {0:'deeppink', 1:'green'}
    
    df_pred = pd.DataFrame({'actual_class': y_test, 'pred_class': pred_class, 'pred_prob': pred_prob})
    
    sns.kdeplot(data=df_pred, x='pred_prob', hue='pred_class', fill=True, legend=False, palette=palette, ax=ax)
    
    ax.set_facecolor('#274472')
    
    ax.set_xlabel('Probability of event', fontsize=14, color='white')
    
    ax.set_ylabel('Probability Density', fontsize=14)
    
    ax.set_title(f'Probability distribution ({dataset})', fontsize=16)
    
    ax.tick_params(axis='both', which='major', labelsize=12)
    
    ax.grid(False)



def _plot_feature_imp(df_features, ax):
    
    sns.barplot(x='feature', y='importance', data=df_features, palette='RdBu_r', ax=ax) 
    
    ax.tick_params(axis='x', rotation=90)
    
    ax.set_frame_on(False)
    
    ax.set_xlabel('Feature', fontsize=14)
    
    ax.set_ylabel('Feature importance', fontsize=14)
    
    ax.set_title('Features importance', fontsize=16)



def _plot_lift_curve(y_test, pred_prob, ax=None, output="plot", dataset="test"):
    test_set = pd.DataFrame({"target" : y_test, "probability" : pred_prob})
    
    test_set = test_set.sort_values(by="probability", ascending=False)
    
    test_set["decile"] = np.linspace(1, 11, len(test_set), False, dtype=int)

    df_gain = (
        test_set
        .groupby("decile")
        .agg(
            count = ("target", "count"),
            events = ("target", sum) 
        )
        .assign(pct_events = lambda X: X["events"]/X["events"].sum()*100)
        .assign(gain = lambda X: X["pct_events"].cumsum())
        .reset_index()
        .assign(decile_pct = lambda X: X["decile"]*10)
        .assign(lift = lambda X: X["gain"]/X["decile_pct"])
    )

    if output != "plot":
        return df_gain
    else:
        if not ax:
            fig, ax = plt.subplots(figsize=(8,6))

        x = np.array(df_gain.decile_pct)
        y = np.array(df_gain.lift)

        X_Y_Spline = make_interp_spline(x, y)

        X_ = np.linspace(x.min(), x.max(), 500)
        Y_ = X_Y_Spline(X_)

        ax.plot(X_, Y_,color='orange', lw=5, label="Lift curve")
        ax.plot(X_, [1 for x in Y_], ls="--", color='black', lw=3, label="Baseline")

        ax.set_xticklabels([f"{x:.0f}%" for x in ax.get_xticks()])

        ax.legend(loc="center", bbox_to_anchor=(0.5, -0.2), frameon=False, ncol=2)

        ax.set_xlabel("Percentage of sample", labelpad=10, fontsize=14)
        ax.set_ylabel("Lift", labelpad=10, fontsize=14)

        ax.set_title(f"Lift curve ({dataset})", fontsize=16)

        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_visible(False)



